/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "mathlib.h"

/* swaps line A with line B */
static void changeline ( int n, int m, double matrix[n][m], int a, int b ) {
	int i;
	double d;
	for ( i = 0; i < n; i++ ) {
		d = matrix[a][i];
		matrix[a][i] = matrix[b][i];
		matrix[b][i] = d;
	}
}

/* solve linear system using Gauss elimination */
int gausselim ( int n, double A[n][n], double B[n], double x[n] ) {
	int i, j, k;
	double factor;

	for ( j = 0; j < n; j++ ) { // columns
		for ( i = n-1; i > j; i-- ) {
			factor = A[i][j]/A[j][j];
			for ( k = 0; k < n; k++ )
			    A[i][k] -= factor*A[j][k];
			B[i] -= factor*B[j];
		}
		/* is system ok? */
		if ( A[j+1][j+1] == 0.0 && j+1 < n) {
			for ( k = j+2; k < n; k++ )
				if ( A[k][j+1] != 0 )
					break;
			if ( k == n || j+2 == n ) /* undetermined */
				return -1;
			changeline( n, n, A, j+1, k );
			changeline( n, 1, B, j+1, k );
		}
	}

	/* solve system */
	for ( i = n-1; i >= 0; i-- ) {
		x[i] = B[i];
		for ( j = i+1; j < n; j++ )
			x[i] -= (A[i][j]*x[j]);
		x[i] /= A[i][i];
	}
	return 0;
}
