/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <stddef.h>
#include <sys/types.h>
#include <dlfcn.h>
#include <malloc.h>
#include <ctype.h>
#include "unlimited.h"

#define ALLOCERROR	"Could not allocate memory"
#define DATAERROR	"File damaged: "

#define BUFFERSIZE	512
#define LSQN		3

TERRAIN_UL	*firstterrain;
OBJECT_UL	*firstobj;

sInt findTrack ( uByte *trackname ) {
	DIR			*tracks;
	uByte			buffer[BUFFERSIZE];
	struct 			dirent *de;
	sInt 			i = 1;
	
	sprintf( buffer, "%s%s", ULDIR, TRACKDIR );
	if ( !(tracks = opendir( buffer )) ) {
		fprintf( stderr, 
		  "You don't have a tracks directory or I can't access it!\n" );
		return FATALERROR;
	}
	sprintf( buffer, "%s%s", trackname, TRACKEXTENSION );
	while ( (de = readdir( tracks )) != NULL && i != 0 )
             i = strcmp( de->d_name, buffer );
	closedir( tracks );
	if ( i == 0 )
		return OK;
	return ERROR;
}

/* filenames passed as arguments should never have extension. */
TRACK_UL *readTrack ( uByte *filename ) {
	FILE 			*fp, *scores;
	TRACK_UL		*track;
	TERRAIN_UL		*terrain;
	OBJECT_UL		*obj;
#ifdef USE_GETLINE
	uByte			*t, *buffer;
#else
	uByte			*t, buffer[BUFFERSIZE];
#endif
	sInt		lines = 0, size = 0, i = 0, j, k, n;
	dFloat		version, d;
	VECTOR_UL	p1, p2;
#ifdef PART_OF_OLD_CODE_TO_SMOOTH_TERRAIN_KEEP_BY_NOW
	double f[LSQN+1][LSQN+1], interp[tilex][tiley], x[LSQN+1], y[LSQN+1];
	double d1, d2;
#endif
	track = (TRACK_UL *)malloc(sizeof(TRACK_UL));
	if ( !track ) {
		fprintf( stderr, "%s: %s readTrack\n", ALLOCERROR, filename );
		return NULL;
	}
	
	sprintf( buffer, "%s%s%s%s", ULDIR, TRACKDIR, filename, TRACKEXTENSION );
	if ( !(fp = fopen( buffer, "r" )) ) { 
		fprintf( stderr, "Could not open file %s\n", filename );
		return NULL;
	}
	
	/* check for header */
	lines += readString( fp, buffer, &size );
	if ( !( t = strstr( buffer, "ULTRACK" )) ) {
		fprintf( stderr, "%s is not a valid track\n", filename );
		free( track );		fclose( fp );
#ifdef USE_GETLINE
		free( buffer );
#endif
		return NULL;
	}
	sscanf( t, "%lf", &version );
	if ( version != 1.0 )
		fprintf( stderr, "%s version is unknown, trying to read anyway\n",
				 filename );		
	
	/* read track name */
	lines += readString( fp, buffer, &size );
	t = buffer;
	while ( isspace(*t) ) t++;
	track->name = strdup( t );
	
	/* read track author */
	lines += readString( fp, buffer, &size );
	t = buffer;
	while ( isspace(*t) ) t++;
	track->author = strdup( t );
	
	/* read track creation date */
	lines += readString( fp, buffer, &size );
	t = buffer;
	while ( isspace(*t) ) t++;
	track->date = strdup( t );
	
	/* read track comment */
	lines += readString( fp, buffer, &size );
	t = buffer;
	while ( isspace(*t) ) t++;
	track->comment = (uByte *)malloc(strlen(buffer)*sizeof(uByte));
	if ( !track->comment ) {
		fprintf( stderr, "%s: while reading %s's comment\n", ALLOCERROR,
				filename );
		free( track->name );		free( track->author );
		free( track->date );		free( track );
		fclose( fp );
#ifdef USE_GETLINE
		free( buffer );
#endif
		return NULL;
	}

	while ( *t != '\0' ) {
		if ( *t == '\\' && *(t+1) == 'n' ) {
			track->comment[i++] = '\n';
			t += 2;
		}
		else 	track->comment[i++] = *t++;
	}
	track->comment[i] = '\0';
	realloc( track->comment, i+1 );

	/* read track laps */
	lines += readString( fp, buffer, &size );
	t = buffer;
	sscanf( t, "%d", &track->totallap );
	
	/* read track weather */
	lines += readString( fp, buffer, &size );
	t = buffer;
	sscanf( t, "%d", &track->weather );

	/* read useflags */
	lines += readString( fp, buffer, &size );
	if ( strstr( buffer, "True" ) )
		track->useflag = True;
	else
		track->useflag = False;

	/* read lap length */
	lines += readString( fp, buffer, &size );
	t = buffer;
	sscanf( t, "%f", &track->length );

	/* done with header, let's read the terrains */
	lines += readString( fp, buffer, &size );
	if ( !( t = strstr( buffer, "terrains" ) ) ) {
		fprintf( stderr, "%s%s: terrains not found, it should be at line %d",
				 DATAERROR, filename, lines );
		free( track->name );		free( track->author );
		free( track->date );		free( track->comment );
		free( track );			fclose( fp );
#ifdef USE_GETLINE
		free( buffer );
#endif
		return NULL;
	}
	sscanf( t, "%d", &i );

	lines += readString( fp, buffer, &size );
	t = buffer;
	for ( j = 0; j < i; j++ ) {
		sscanf( t, "%d:%n", &k, &n );
		t += n;
		readTerrain( k, t );
	}

	/* read the landscape */
	lines += readString( fp, buffer, &size );
	t = buffer;
	if ( readGeometry( buffer, track->geom ) == FATALERROR ) {
		fprintf( stderr, "readGeometry died reading %s, so do we reading %s\n",
				buffer, filename );
		free( track->name );		free( track->author );
		free( track->date );		free( track->comment );
		free( track );			fclose( fp );
		closeTerrains();
#ifdef USE_GETLINE
		free( buffer );
#endif
		return NULL;
	}

	/* come on, hurry up! read the tracks now! slow computer! */
	lines += readString( fp, buffer, &size );
	t = buffer;
	if ( !( t = strstr( buffer, "tracks" ) ) ) {
		fprintf( stderr, "%s%s: tracks not found, it should be at line %d",
				 DATAERROR, filename, lines );
		free( track->name );		free( track->author );
		free( track->date );		free( track->comment );
		freeDoodad( track->geom );	free( track );
		fclose( fp );			closeTerrains();
#ifdef USE_GETLINE
		free( buffer );
#endif
		return NULL;
	}
	sscanf( t, "%d", &i );
	if ( i == 0 ) /* no tracks, let's jump */
		goto objects;
	
	if ( !(track->ways = (WAY_UL *)malloc(i*sizeof(WAY_UL))) ) {
		fprintf( stderr, "%s: %s's terrains", ALLOCERROR, filename );
		free( track->name );		free( track->author );
		free( track->date );		free( track->comment );
		freeDoodad( track->geom );	free( track );
		fclose( fp );			closeTerrains();
#ifdef USE_GETLINE
		free( buffer );
#endif
		return NULL;
	}

	for ( j = 0; j < i; j++ ) {
		lines += readString( fp, buffer, &size );
		t = buffer;
		/* point */
		sscanf( t, "%lf %lf %lf%n", &track->ways[j].point.x,
			 &track->ways[j].point.y, &track->ways[j].point.z, &n );
		t += n;
		while ( isspace(*t) ) t++;
		
		/* is checkpoint? */
		if ( *t == 'p' ) {
			track->ways[j].checkpoint = True;
			t++;
		}
		else
			track->ways[j].checkpoint = False;
		
		if ( sscanf( t, "%lf %lf %lf %lf %lf %lf %lf %lf ", 
			&track->ways[j].width, &d, &track->ways[j].normal.x,
			&track->ways[j].normal.y, &track->ways[j].normal.z,
			&track->ways[j].tan.x, &track->ways[j].tan.y,
			&track->ways[j].tan.z ) != 8 ) {
			fprintf( stderr, "%s: %s, incorrect track point at line %d\n",
				DATAERROR, filename, lines );
			free( track->name );		free( track->author );
			free( track->date );		free( track->comment );
			free( track->ways );		closeTerrains();
			free( track );			fclose( fp );
#ifdef USE_GETLINE
			free( buffer );
#endif
			return NULL;
		}
		
		for ( terrain = firstterrain; terrain != NULL; terrain = terrain->next )
			if ( terrain->id == n )
				track->ways[j].terrain = terrain;
	}

objects:
	/* last, but not least, the objects... */
	while ( !feof( fp ) ) {
		lines += readString( fp, buffer, &size );
		t = strchr( buffer, ':' );
		*t++ = '\0';
		sscanf( t, "%d %lf %lf %lf %lf %lf %lf", &i, &p1.x, &p1.y, &p1.z,
				&p2.x, &p2.y, &p1.z );
#ifndef SERVER /*** these are buggy! */
		obj = readObject( 0, buffer, &p1, &p2 );
#else
/*		need to add obj as server. id= 0x0000+...? */
#endif
		
		for ( j = 1; j < i; j++ ) { /* copies */
			lines += readString( fp, buffer, &size );
/**** check sscanf here */
			sscanf( buffer, "%d %lf %lf %lf %lf %lf %lf", &i, &p1.x, &p1.y,
				&p1.z, &p2.x, &p2.y, &p1.z );
			copyObject( obj, &p1, &p2, d );
		}
	}
	fclose( fp );
	
#ifdef PART_OF_OLD_CODE_TO_SMOOTH_TERRAIN_KEEP_BY_NOW
	for ( i = 0; i <= LSQN; i++ )
		x[i] = y[j] = TILESIZE*i;
	
	/**** TODO: handle the borders of the map, they are not smoothed by
	now */
	for ( i = 0; i < track->width-2; i++ )
		for ( j = 0; j < track->length-2; i++ ) {

			/* fill f with the points we have */
			for ( k = 0; k < LSQN; k++ )
				for ( l = 0; l < LSQN; l++ )
					f[k][l] = 
					     track->tiles[i+k][j+l].heights[3];
					
			for ( k = 0; k < LSQN; k++ )
				f[LSQN][k] = 
					track->tiles[i+k][j+LSQN-1].heights[2];
			for ( k = 0; k < LSQN; k++ )
				f[k][LSQN] = 
					track->tiles[i+LSQN-1][j+k].heights[0];
			f[LSQN][LSQN] = 
				track->tiles[i+LSQN-1][j+LSQN-1].heights[1];

			/* call lsq and save output in interp */
			lsq(LSQN, LSQN, x, y, f, TILESIZE*LSQ/3, TILESIZE*LSQ/3,
				2*TILESIZE*LSQ/3, 2*TILESIZE*LSQ/3,
				prefs->tilex, prefs->tiley, 
				track->tile[i+1][j+1].groundg );
		}
#endif

	/* now get scores */
	sprintf( buffer, "%s%s%s%s", ULDIR, TRACKDIR, filename, SCOREEXTENSION );
	scores = fopen( buffer, "r" );
	if ( !scores ) {
		fprintf( stderr, "%s%s not found, creating one...\n",
				 filename, SCOREEXTENSION );
		if ( !(scores = fopen( buffer, "w" ) ) ) {
			fprintf( stderr, "Couldn't create %s%s...\n",
				 filename, SCOREEXTENSION );
			return track;
		}
		for ( i = 0; i < 5; i++ )
			fprintf( scores, "No one\nVoid\n0 0 0 0\n" );

#ifdef USE_GETLINE
		free( buffer );
#endif
		return track;
	}
	
	for ( i = 0; i < 5; i++ ) {
		track->besttimes[i].position = i+1;
		lines += readString( fp, buffer, &size );
		t = buffer;
		while ( isspace(*t) ) t++;
		track->besttimes[i].driver = strdup( buffer );
		lines += readString( fp, buffer, &size );
		t = buffer;
		while ( isspace(*t) ) t++;
		track->besttimes[i].model = strdup( buffer );
		lines += readString( fp, buffer, &size );
		fscanf( scores, "%d %hd %d %d %lu", 
				&track->besttimes[i].weather,
				&track->besttimes[i].laptime.h,
				&track->besttimes[i].laptime.min,
				&track->besttimes[i].laptime.sec,
				&track->besttimes[i].laptime.usec );
	}	  
	fclose( scores ); 
#ifdef USE_GETLINE
	free( buffer );
#endif
	return track;
}

sInt saveScores ( TRACK_UL *track ) {
	FILE 			*scores;
	sInt			i;
	uByte			buffer[BUFFERSIZE];

	sprintf( buffer, "%s%s%s%s", ULDIR, TRACKDIR, track->name, SCOREEXTENSION );
	if ( (scores = fopen( buffer, "w")) == NULL ) {
		fprintf( stderr, "Could not save %s scores", track->name );
		return ERROR;
	}
	for ( i = 0; i < 5; i++ ) {
		fprintf( scores, "%s\n", track->besttimes[i].driver );
		fprintf( scores, "%s\n", track->besttimes[i].model );
		fprintf( scores, "%d %d %d %d %ld\n",
			(WEATHER_UL) track->besttimes[i].weather,
			track->besttimes[i].laptime.h,
			track->besttimes[i].laptime.min,
			track->besttimes[i].laptime.sec,
			track->besttimes[i].laptime.usec );
	}
	fclose( scores );
	return OK;
}

void closeTrack ( TRACK_UL *track ) {
	sInt			i;

	free( track->name );
	free( track->author );
	free( track->date );
	free( track->comment );
	free( track->ways );
	freeDoodad( track->geom );
	
	for ( i = 0; i < 5; i++ ) {
		free( track->besttimes[i].driver );
		free( track->besttimes[i].model );
	}
	free( track );
	closeTerrains();
}

sInt readTerrain ( sInt id, uByte *name ) {
	sInt			size = 0, lines = 0, red, green, blue;
#ifdef USE_GETLINE
	uByte			*t, *buffer;
#else
	uByte			*t, buffer[BUFFERSIZE];
#endif
	sByte			error[] = "ERROR: terrains.ul corrupted";
	FILE *fp;
	TERRAIN_UL *newt, *tmp;
	
	/* check if terrain is already loaded */
	for ( tmp = firstterrain; tmp != NULL; tmp = tmp->next )
		if ( tmp->id == id )
			return OK;

	sprintf( buffer, "%s%s", ULDIR, TERRAININIT );
	if ( (fp = fopen( TERRAININIT, "r" )) == NULL ) {
		fprintf( stderr, "FATAL ERROR: Could not open terrain files\n" );
		return FATALERROR;
	}

	if ( (newt = (TERRAIN_UL *)malloc(sizeof(TERRAIN_UL))) == NULL ) {
		fprintf( stderr, "%s: terrain\n", ALLOCERROR );
		return FATALERROR;
	}

	while ( !feof( fp ) ) {
		lines += readString( fp, buffer, &size );
		t = buffer;
		while ( isspace(*t) ) t++;

		/* is it the one we want? */
		if ( !strcmp( t, name ) ) {
			/* no, find next ; */
			lines += readString( fp, buffer, &size );
			t = buffer;
			while ( isspace(*t) ) t++;
			if ( t[0] == 'e' && t[1] == 'o' && t[0] == 't' )
				continue;
		}
		
		/* yes, go on */
		newt->name = strdup(name);
		newt->id = id;
		
		/* read texture map file */
		lines += readString( fp, buffer, &size );
		t = buffer;
/****** open texture, etc */
		
		/* read sound */
		lines += readString( fp, buffer, &size );
		while ( isspace( *t ) )
			t++;
#ifdef CLIENT
		newt->sound = strdup( buffer );
#endif

		/* color */
		lines += readString( fp, buffer, &size );
		if ( (sscanf( buffer, "%d %d %d", &red, &green, &blue )) != 3 ) {
			fprintf( stderr, "%s: rgb at line %d; we made it grey\n",
					error, lines );
			red = green = blue = 128;
		}
#ifdef CLIENT
		newt->color = RGBAtoColor( red, green, blue, 0 );
#endif
		lines += readString( fp, buffer, &size );
		if ( !sscanf( buffer, "%lf", &newt->maxspeed ) ) {
			fprintf( stderr, "%s: maxspeed at line %d. Assuming 1.0\n",
					error, lines );
			newt->maxspeed = 1.0;
		}

		lines += readString( fp, buffer, &size );
		if ( sscanf( buffer, "%lf %lf %lf %lf", &newt->friction[0],
			&newt->friction[1], &newt->friction[2],
			&newt->friction[3] ) != 4 ) {
			fprintf( stderr, "%s: frictions at line %d. Assuming 0.2\n",
					error, lines );
			newt->friction[0] = newt->friction[1] = 
				newt->friction[2] = newt->friction[3] = 0.2;
		}

		lines += readString( fp, buffer, &size );
		t = buffer;
		while ( isspace(*t) ) t++;
		if ( t[0] != 'e' || t[1] != 'o' || t[0] != 't' ) {
			fprintf( stderr, "%s: \'eot\' not found at line %d;\nignoring and going on (this may lead to unpredictable behavior)\n",
					 error, lines );
		}
	}

	if ( firstterrain == NULL )
		firstterrain = newt;
	else
		for ( tmp = firstterrain; tmp->next != NULL; tmp = newt->next )
			tmp->next = newt;
	newt->next = NULL;
#ifdef USE_GETLINE
	free( buffer );
#endif
	return OK;
}

void closeTerrains ( void ) {
	TERRAIN_UL *tmp, *next;
	
	if ( !firstterrain )
		return;

	tmp = firstterrain;
	next = firstterrain;      	
	while ( next ) {
		next = next->next;	
		free( tmp->name );
#ifdef CLIENT
		free( tmp->sound );
/**** free( texture ) */
#endif
		free( tmp );
		tmp = next;
	}
}

sInt findClient ( uByte *clientname ) {
	DIR 			*clientdir;
	uByte 			buffer[BUFFERSIZE], buffer2[BUFFERSIZE];
	struct			dirent *de;
	sInt			i = 1;

	sprintf( buffer, "%s%s", ULDIR, CLIENTDIR );
	if ( !(clientdir = opendir( buffer )) ) {
		fprintf( stderr, 
		  "You don't have a cli directory or I can't access it!\n" );
		return FATALERROR;
	}
	sprintf( buffer, "%s%s", clientname, ".geom" );
	sprintf( buffer, "%s%s", clientname, ".so" );
	while ( (de = readdir( clientdir )) != NULL && i != 0 ) {
		if ( !strcmp( de->d_name, buffer ) )
			i += 1;
		if ( !strcmp( de->d_name, buffer2 ) )
			i += 2;
	}
	closedir( clientdir );
	if ( i == 0 )
		return OK;
	if ( i & 0x01 ) /* we couldn't find the geometry */
		fprintf( stderr, "It seems that you don't have %s. We'll use the default geometry\n",
			buffer );
	if ( i & 0x02 ) { /* we couldn't find the shared library */
		fprintf( stderr, "It seems that you don't have %s. No way to proceed.\n",
			buffer2 );
		return FATALERROR;
	}
	return ERROR;
}

#ifndef SERVER
CLIENT_UL *readClient ( uByte *name ) {
	uByte			buffer[BUFFERSIZE];
	CLIENT_UL 		*cli;
	const sByte		*error;

	cli = (CLIENT_UL *) malloc (sizeof(CLIENT_UL));
	if ( !cli ) {
		fprintf( stderr, "%s readClient1", ALLOCERROR );
		return NULL;
	}

	sprintf( buffer, "%s%s%s%s", ULDIR, CLIENTDIR, name, ".geom" );
	if ( readGeometry( buffer, cli->geom ) == FATALERROR ) {
		fprintf( stderr, "ERROR: %s. Could not read cli's geometry, trying default.\n",
			DATAERROR  );
		sprintf( buffer, "%s%s%s", ULDIR, CLIENTDIR, "default.geom" );
		if ( readGeometry( buffer, cli->geom ) == FATALERROR ) {
			fprintf( stderr, "FATAL ERROR: %s. Could not default geometry %s!\n",
				DATAERROR, buffer );
			free( cli );
			return NULL;
		}
	}
		
	sprintf( buffer, "%s%s%s%s", ULDIR, CLIENTDIR, name, ".so" );
	cli->lib = dlopen( buffer, RTLD_LAZY );
	if ( !cli->lib ) {
		fprintf( stderr, "Could not find %s%s. This error shouldn't happen here. This is not a joke.\n", 
				name, ".so" );
		free( cli->model );	freeDoodad( cli->geom );
		free( cli );
		return NULL;
	}
	cli->init = dlsym( cli->lib, "initClient" );
	if ( (error = dlerror()) != NULL )
		goto liberror;

	cli->finish = dlsym( cli->lib, "closeClient" );
	if ( (error = dlerror()) != NULL )
		goto liberror;

	cli->event = dlsym( cli->lib, "handleEvent" );
	if ( (error = dlerror()) != NULL )
		goto liberror;
	
	cli->loop = dlsym( cli->lib, "mainLoop" );
	if ( (error = dlerror()) != NULL ) {
liberror:
		fprintf( stderr, "FATAL: Couldn't open a necessary function in %s.so. libdl said that:\n%s\n",
				cli->model, error );
		free( cli->model );	dlclose( cli->lib );
		freeDoodad( cli->geom );	free( cli );
		return NULL;
	}
	/* these functions are optional */
	cli->message = dlsym( cli->lib, "handleSpecialMessage" );

#ifdef OLD_CODE_TO_READ_WHEELS
	lines += readString( fp, buffer, &size );
	if ( !strstr( buffer, "wheels:" ) ) {
		fprintf( stderr, "%s%s", DATAERROR, name );
		free( cli->model );	dlclose( cli->lib );
		freeDoodad( cli->dood );
		free( cli );		fclose( fp );
#ifdef USE_GETLINE
		free( buffer );
#endif
		return NULL;
	}
	t = strchr( buffer, ':' ) + 1;
	sscanf( t, "%d", &i );
	cli->wheels = (WHEEL_UL *)malloc(i*sizeof(WHEEL_UL));
	for ( j = 0; j < i; j++ ) {
		lines += readString( fp, buffer, &size );
		sscanf( buffer, "%f %f %f", cli->wheels[i].maxsteer,
				cli->wheels[i].radius, cli->wheels[i].width );
		cli->wheels[i].state = NORMAL;
		cli->wheels[i].wearing = 1000000;
		cli->wheels[i].circ = 2*PI*cli->wheels[i].radius;
		cli->wheels[i].rotation = 0;
		cli->wheels[i].angle = 0;
		zeroV( &cli->wheels[i].normal );
		cli->wheels[i].temperature = general.temperature;
	}
#endif /*OLD_CODE */

#ifdef USE_GETLINE
	free( buffer );
#endif
	return cli;

}
#endif /* SERVER */

sInt intnofunc ( OBJECT_UL *obj ) {
	return OK;
}

void voidnofunc ( OBJECT_UL *obj ) {
}

static uShort objn = 1;

#ifndef SERVER
OBJECT_UL *readObject ( uLong clientid, uByte *name,
		 VECTOR_UL *pos, VECTOR_UL *dir ) {
#else
OBJECT_UL *readObject ( uLong id, uByte *name, VECTOR_UL *pos, VECTOR_UL *dir ) {
#endif
	uByte 			buffer[BUFFERSIZE];
	OBJECT_UL 		*obj, *temp;
	const sByte		*error;
	Bool			b;
	
	temp = firstobj;

	sprintf( buffer, "%s%s%s%s", ULDIR, OBJECTDIR, name, ".geom" );
	
	if ( !(obj = (OBJECT_UL *)malloc(sizeof(OBJECT_UL))) ) {
		fprintf( stderr, "%s: readObject", ALLOCERROR );
		return NULL;
	}

	if ( readGeometry( buffer, obj->geom ) == FATALERROR ) {
		free( obj );
		return NULL;
	}
	copyV( &obj->geom->world, pos );
	copyV( &obj->geom->dir, dir );

	if ( name )
		obj->name = strdup( name );

#ifndef SERVER
	if ( general.useobjso ) {
		sprintf( buffer, "%s%s%s%s", ULDIR, CLIENTDIR, name, ".so" );
		obj->lib = dlopen( buffer, RTLD_LAZY );
		if ( !obj->lib ) {
			obj->init = intnofunc;
			obj->finish = voidnofunc;
			goto nolib;
		}
		
		obj->init = dlsym( obj->lib, "initObject" );
		if ( (error = dlerror()) != NULL ) {
			fprintf( stderr, "Couldn't open initObject() in %s.so. libdl said that:\n%s\n",
				obj->name, error );
			free( obj->name );	dlclose( obj->lib );
			freeDoodad( obj->geom ); free( obj );
			return NULL;
		}

		obj->finish = dlsym( obj->lib, "closeObject" );
		if ( (error = dlerror()) != NULL ) {
			fprintf( stderr, "Couldn't open closeObject() in %s.so. libdl said that:\n%s\n",
				obj->name, error );
			free( obj->name );	dlclose( obj->lib );
			freeDoodad( obj->geom ); free( obj );
			return NULL;
		}

		obj->brain = dlsym( obj->lib, "simulateObject" );
		if ( (error = dlerror()) != NULL ) {
			fprintf( stderr, "Couldn't open simulateObject() in %s.so. libdl said that:\n%s\n",
				obj->name, error );
			free( obj->name );	dlclose( obj->lib );
			freeDoodad( obj->geom ); free( obj );
			return NULL;
		}
	}

	if ( obj->init( obj ) != OK ) { /* it couldn't initialize */
		freeObject( obj );
		return NULL;
	}
nolib:
	do { 
	/* this loop is a waste of time in normal games with few objects
	  created, but things really will mess up if the counter reaches
	  zero again */
		b = True;
		obj->id = clientid + ( objn < 0xFFFF ? objn++ : 1 );
		for ( temp = firstobj; temp != NULL ; temp = temp->next )
			if ( temp->id == obj->id )
				b = False;
	} while ( b == False );

#else /* SERVER */
	obj->id = id;
#endif

	/* we only add it to the list if everything's ok */
	if ( firstobj == NULL )
		firstobj = obj;
	else {
		temp = firstobj;
		while ( temp->next != NULL )
			temp = temp->next;
		temp->next = obj;
	}

	return obj;
}

OBJECT_UL *copyObject( OBJECT_UL *obj, VECTOR_UL *dir, VECTOR_UL *pos, dFloat
		angle ) {
	OBJECT_UL		*new, *temp;
	
	if ( (new = (OBJECT_UL *)malloc(sizeof(OBJECT_UL))) == NULL ) {
		fprintf( stderr, "%s: copyObject", ALLOCERROR );
		return NULL;
	}
	memcpy( new, obj, sizeof(OBJECT_UL) );
	copyV( &new->direction, dir );
	copyV( &new->position, pos );
	
	if ( firstobj == NULL )
		firstobj = new;
	else {
		temp = firstobj;
		while ( temp->next != NULL )
			temp = temp->next;
		temp->next = new;
	}
	return new;
}

void freeObject ( OBJECT_UL *obj ) {
	OBJECT_UL		*tmp, *prev;
	
	if (!obj )
		return;
	
	prev = tmp = firstobj;

	while ( tmp ) {
		if ( tmp == obj )
			break;
		prev = tmp;
		tmp = tmp->next;
	}
	
	if ( tmp == NULL )
		fprintf( stderr, "Object is not in list! Freeing anyway\n" );
	else {
		if ( tmp == firstobj )
			firstobj = tmp->next;
		else 
			prev->next = tmp->next;
	}

	obj->finish( obj );
	freeDoodad( obj->geom );
	free( obj->name );
	dlclose( obj->lib );
	free( obj );
}
