/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "unlimited.h"

static struct timeval	starttime;

/* timeval_subtract adapted from libc's info documentation */
TIMESTAMP_UL timeval_subtract ( struct timeval *x, struct timeval *y, 
	struct timeval *result ) {
	int		nsec;
	struct timeval	temp;

	/* Perform the carry for the later subtraction by updating Y. */
	if (x->tv_usec < y->tv_usec) {
		nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		temp.tv_sec = y->tv_usec - 1000000 * nsec;
		temp.tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		nsec = (y->tv_usec - x->tv_usec) / 1000000;
		temp.tv_sec = y->tv_usec + 1000000 * nsec;
		temp.tv_sec -= nsec;
	}

	/* Compute the time remaining to wait.
	   `tv_usec' is certainly positive. */
	if ( result ) {
		result->tv_sec = x->tv_sec - temp.tv_sec;
		result->tv_usec = x->tv_usec - temp.tv_sec;
		return (result->tv_sec*1000+(result->tv_usec)/1000);
	}

	/* this will blow every 49 days or so. By now this is not a problem,
	but it'd be nice to solve it. I guess the best way to do is calling
	inittime() again, and resynchronize the clients. */
	return (TIMESTAMP_UL)((x->tv_sec - y->tv_sec)*1000+(x->tv_usec - y->tv_usec)/1000);
}

/* the struct timeval is here for completeness. The network communications will
not use it. */
TIMESTAMP_UL timestamp ( struct timeval *stamp ) {
	struct timeval	tv;
	gettimeofday( &tv, NULL );
	return timeval_subtract( &tv, &starttime, stamp );
}

void inittime ( void ) {
	gettimeofday( &starttime, NULL );
}

TIME_UL elapsedtime ( void ) {
	struct timeval 	tv;
	TIME_UL		t;

	gettimeofday( &tv, NULL );
	timeval_subtract( &tv, &starttime, &tv );
	t.usec = tv.tv_usec;
	t.h = tv.tv_sec/3600;
	t.min = tv.tv_sec/60 - t.h*60;
	t.sec = tv.tv_sec - (t.min*60 + t.h*3600);
	
	return t;
}

TIMESTAMP_UL timetots ( unsigned long minutes, unsigned long seconds, 
		unsigned long usec ) {
	return (TIMESTAMP_UL)(60*minutes+seconds)*1000+usec/1000;
}
