/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#ifdef CLIENT

#define	NOMEM		"Could not allocate memory"

#include "ulengine.h"
#include <malloc.h>
#include <string.h>

#ifdef USE_JPEG
# include <jpeglib.h>
# include <setjmp.h>
#endif
#ifdef USE_GIF

#endif
#ifdef USE_PNG

#endif
#ifdef USE_TIFF
# include <tiffio.h>
#endif


#ifdef USE_JPEG
/* code to load jpegs is adapted from jpeglib's example.c */

struct my_jpeg_error_mgr {
	struct		jpeg_error_mgr pub;
	jmp_buf		setjmp_buffer;
};
typedef struct my_error_mgr * my_error_ptr;

static void my_jpeg_error_exit ( j_common_ptr cinfo ) {
	my_error_ptr myerr = (my_error_ptr) cinfo->err;
	(*cinfo->err->output_message) (cinfo);
	longjmp(myerr->setjmp_buffer, 1);
}

static IMAGE_UL *loadjpeg ( const uByte *filename ) {
	IMAGE_UL		*image;
	JSAMPARRAY 		buffer;
	FILE			*fp;
	sInt			row_stride;
	struct my_jpeg_error_mgr	jerr;
	struct jpeg_decompress_struct	cinfo;
	uByte			*r, *g, *b;

	if ( !(image = (IMAGE_UL *)malloc(sizeof(IMAGE_UL))) ) {
		fprintf( stderr, "%s: image\n", NOMEM );
		return NULL;
	}
	if ( !(fp = fopen( fp, "rb" )) ) {
		fprintf( stderr, "Could not find %s\n", filename );
		free( image );
		return NULL;
	}
	image->name = strdup( filename );

	cinfo.err = jpeg_std_error( &jerr.pub );
	jerr.pub.error_exit = my_jpeg_error_exit;

	if ( setjmp( jerr.setjmp_buffer ) ) {
		jpeg_destroy_decompress( &cinfo );
		fclose( fp );
		if ( image ) {
			if ( image->data ) {
				for ( i = 0; image->data[i] != NULL; i++ );
					free( image->data[i] );
				free( image.data );
			}
			if ( image->name ) 
				free( image->name );
			free( image );
		}
		return NULL;
	}

	jpeg_create_decompress( &cinfo );
	jpeg_stdio_src( &cinfo, fp );
	jpeg_read_header( &cinfo, TRUE );
	
	image->height = cinfo.image_height;
	image->width = cinfo.image_width;
	if ( cinfo.jpeg_color_space != JCS_RGB ) {
		fprintf( stderr, "Image %s is not saved in RGB mode. Maybe jpeglib\nwon't be able to convert it.\n",
				filename );
	}
	cinfo.out_color_space = JCS_RGB;

	if ( !(image->data = (ggi_color **)malloc(image->height*sizeof(ggi_color *)))) {
		fprintf( stderr, "%s: image, data 1\n", NOMEM );
		free( image );
		jpeg_destroy_decompress( &cinfo );
		fclose( fp );
		return NULL;
	}
	for ( i = 0; i < image->height; i++ )
		if ( !(image->data[i] = (ggi_color *)malloc(image->width*sizeof(ggi_color)))) {
			fprintf( stderr, "%s: image, data 2\n", NOMEM );
			for ( i--, i >= 0; i-- ) 
				free( image->data[i] );
			free( image->data );	free( image );
			jpeg_destroy_decompress( &cinfo );
			fclose( fp );
			return NULL;
		}

	jpeg_start_decompress(&cinfo);
	row_stride = cinfo.output_width * cinfo.output_components;
	buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

	for ( j = 0; cinfo.output_scanline < cinfo.output_height; j++ ) {
		jpeg_read_scanlines( &cinfo, buffer, 1 );
		/* I assume output in 24 bits */
		r = &buffer[0];		g = r + 1;	b = g + 1;
		for ( i = 0; i < cinfo.output_width; i++, r += 3, g += 3, b += 3 )
			image->data[j][i] = RGBAtoColor( *r, *g, *b, 0 );
	}

	jpeg_finish_decompress( &cinfo );
	jpeg_destroy_decompress( &cinfo );
	fclose( fp );
	return image;
}
#endif

#ifdef USE_GIF
static IMAGE_UL *loadgif ( const uByte *filename ) {
	IMAGE_UL		*image;

	if ( !(image = (IMAGE_UL *)malloc(sizeof(IMAGE_UL))) ) {
		fprintf( stderr, "%s: image\n", NOMEM );
		return NULL;
	}


	return image;
}
#endif

#ifdef USE_PNG
static IMAGE_UL *loadpng ( const uByte *filename ) {
	IMAGE_UL		*image;

	if ( !(image = (IMAGE_UL *)malloc(sizeof(IMAGE_UL))) ) {
		fprintf( stderr, "%s: image\n", NOMEM );
		return NULL;
	}

	return image;
}
#endif

#ifdef USE_TIFF
static IMAGE_UL *loadtiff ( const uByte *filename ) {
	IMAGE_UL		*image;
	TIFF 			*tiffim;
	sInt			i, j;
	uLong			*buffer;
	uint32			t; /* complying with libtiff */

	if ( !(image = (IMAGE_UL *)malloc(sizeof(IMAGE_UL))) ) {
		fprintf( stderr, "%s: image\n", NOMEM );
		return NULL;
	}

	if ( !(tiffim = TIFFOpen( filename, "r" )) ) {
		fprintf( stderr, "Couldn't open %s\n", filename );
		free( image );
		return NULL;
	}

	if ( !(image->name = strdup(filename)) )
		fprintf( stderr, "Can't save image name %s. If Unlimited didn't crash, this is weird.\n", filename );
		
	/* get image specs */
	TIFFGetField( tiffim, TIFFTAG_IMAGEHEIGHT, &t );
	image->height = t - 1;		image->invheight = 1/(fFloat)image->height;
	TIFFGetField( tiffim, TIFFTAG_IMAGEWIDTH, &t );
	image->width = t - 1;		image->invwidth = 1/(fFloat)image->width;

	/* it seems that memory is to be wasted... */
	if ( !(buffer = (uLong)malloc(image->width*image->height*sizeof(Ulong)))) {
		fprintf( stderr, "%s: image, buffer\n", NOMEM );
		free( image );
		TIFFClose( tiffim );
		return NULL;
	}
	/* so, let's do it some more! :) */
	if ( !(image->data = (ggi_color **)malloc(image->height*sizeof(ggi_color *)))) {
		fprintf( stderr, "%s: image, data 1\n", NOMEM );
		free( buffer );		free( image );
		TIFFClose( tiffim );
		return NULL;
	}
	for ( i = 0; i < image->height; i++ )
		if ( !(image->data[i] = (ggi_color *)malloc(image->width*sizeof(ggi_color)))) {
			fprintf( stderr, "%s: image, data 2\n", NOMEM );
			free( buffer );		TIFFClose( tiffim );
			for ( i--, i >= 0; i-- ) 
				free( image->data[i] );
			free( image->data );	free( image );
			return NULL;
		}

	/* read to a friendly RGB format */
	if ( !TIFFReadRGBAImage( tiffim, image->width, image->height,
			buffer, 1 ) ) {
		fprintf( stderr, "Can't read %s\n", filename );
		free( buffer );		TIFFClose( tiffim );
		for ( i = 0; i < image->height; i++ )
			free( image->data[i] );
		free( image->data );	free( image );
		return NULL;
	}

	/* and save in our color structures */
	for ( i = 0; i < image->height; i++ )
		for ( j = 0; j < image->width; i++ )
			/* alpha channel is cleared... */
			image->data[i][j] = RGBAtoColor( TIFFGetR(buffer[i*image->width + x])<<8,
					TIFFGetG(buffer[i*image->width + x])<<8,
					TIFFGetB(buffer[i*image->width + x])<<8, 0 );

	free( buffer );
	TIFFClose( tiffim );
	return image;
}
#endif

IMAGE_UL *loadimage ( const uByte *filename ) {
#ifdef USE_JPG
	if ( strstr( filename, ".jpg" ) || strstr( filename, ".jpeg" ) )
		return loadjpeg( filename );
	else 
#endif
#ifdef USE_GIF
	if ( strstr( filename, ".gif" ) )
		return loadgif( filename );
	else
#endif
#ifdef USE_PNG
	if ( strstr( filename, ".png" ) || strstr( filename, ".pbm" ) )
		return loadpng( filename );
	else 
#endif
#ifdef USE_TIFF
	if ( strstr( filename, ".tif" ) || strstr( filename, ".tiff" ) )
		return loadtiff( filename );
#endif
	fprintf( stderr, "Unknown image format: file %s", filename );
	return NULL;
				
}

void freeImage ( IMAGE_UL *image ) {
	sInt			i;

	if ( image->name )
		free( image->name );
	for ( i = 0; i < image->height; i++ )
		free( image->data[i] );
	free( image->data );
	free( image );
}

/* returns corresponding closest pixel from an image, given (u,v) coordinates,
0<=(u,v)<= 1 */
ggi_color pixelImage ( IMAGE_UL *image, fFloat u, fFloat v ) {
	return image->data[round(v*image->height)][round(u*image->width)];
}

#endif
