/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "ulengine.h"
#include "mathlib.h"
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <malloc.h>

#define ALLOCERROR	"Error allocating"

DOODAD_UL *firstdoodad;

#ifdef CLIENT
POINT_UL *initPoint ( dFloat x, dFloat y, dFloat z, dFloat u, dFloat v,
			dFloat nx, dFloat ny, dFloat nz ) {
#else
POINT_UL *initPoint ( dFloat x, dFloat y, dFloat z, dFloat nx, dFloat ny,
			dFloat nz ) {
#endif
	POINT_UL	*p;

	p = (POINT_UL *)malloc(sizeof(POINT_UL));
	if ( !p ) {
		fprintf( stderr, "\n%s initPoly, 1", ALLOCERROR );
		return NULL;
	}
	p->vect = (VECTOR_UL *)malloc(sizeof(VECTOR_UL));
	if ( !p->vect ) {
		fprintf( stderr, "\n%s initPoly, 1", ALLOCERROR );
		return NULL;
	}
	p->norm = (VECTOR_UL *)malloc(sizeof(VECTOR_UL));
	if ( !p->norm ) {
		fprintf( stderr, "\n%s initPoly, 1", ALLOCERROR );
		return NULL;
	}
	p->vect->x = x;
	p->vect->y = y;
	p->vect->z = z;
#ifdef CLIENT
	p->u = u;
	p->v = v;
#endif
	p->norm->x = nx;
	p->norm->y = ny;
	p->norm->z = nz;
	normalizeV( p->norm );
	return p;
}

/* result = p1 - p2 */
void pointstoV ( POINT_UL *p1, POINT_UL *p2, VECTOR_UL *result ) {
	result->x = p1->vect->x - p2->vect->x;
	result->y = p1->vect->y - p2->vect->y;
	result->z = p1->vect->z - p2->vect->z;
}

/* center of polygon */
void centerP ( POLY_UL *poly, VECTOR_UL *p ) {
	sInt		i;
	zeroV( p );
	for ( i = 0; i < poly->npoints; i++ ) {
		p->x += poly->points[i].vect->x;
		p->y += poly->points[i].vect->y;
		p->z += poly->points[i].vect->z;
	}
	p->x /= (dFloat)poly->npoints;
	p->y /= (dFloat)poly->npoints;
	p->z /= (dFloat)poly->npoints;
}

void centroid ( DOODAD_UL *d, VECTOR_UL *p ) {
	sInt		i;
	VECTOR_UL	q, a, b, area2;
	dFloat		area;

	zeroV( &q );
	for ( i = 0; i < d->npoly; i++ ) {
		centerP( d->polygons[i], &q );
		pointstoV( &d->polygons[i]->points[0],
			&d->polygons[i]->points[1], &a );
		pointstoV( &d->polygons[i]->points[0],
			&d->polygons[i]->points[2], &b );
		crossprodV( &a, &b, &area2 );
		area = fabs( normV( &area2 ) )/2.0;
		scalarMV( &q, area, &q );
		p->x += q.x;
		p->y += q.y;
		p->z += q.z;
	}
	p->x /= (dFloat)d->npoly;
	p->y /= (dFloat)d->npoly;
	p->z /= (dFloat)d->npoly;
}

/**********BUGGY */
/* this function is an adaptation of the code for "An Efficient Bounding
Sphere", by Jack Ritter and Lyle Rains, in Graphic Gems I (see ACKNOWLEDGEMENTS)
The engine's polygon structure allowed to improve the code using pointers,
but it's still very bad because every vertex is checked n times in each loop,
where n is the number of edges this vertex belongs to. I don't see a way to
make it jump (except using an array of used points), but as this code is
seldomly used, let the processor work.*/
SPHERE_UL circSphere ( DOODAD_UL *d ) {
	SPHERE_UL	result;
	sInt		i, j;
	dFloat x, y, z, radsq, a, b, maxdist;
	VECTOR_UL *xmin, *xmax, *ymin, *ymax, *zmin, *zmax, *d1, *d2, min, max;
	
	min.x = min.y = min.z = 100000000.0;
	max.x = max.y = max.z = -100000000.0;
	xmin = ymin = zmin = &min;
	xmax = ymax = zmax = &max;

	for ( i = 0; i < d->npoly; i++ )
		for ( j = 0; j < d->polygons[i]->npoints; j++ ) {
			if ( d->polygons[i]->points[j].vect->x < xmin->x )
				xmin = d->polygons[i]->points[j].vect;
			if ( d->polygons[i]->points[j].vect->x < xmax->x )
				xmax = d->polygons[i]->points[j].vect;
			if ( d->polygons[i]->points[j].vect->y < ymin->y )
				ymin = d->polygons[i]->points[j].vect;
			if ( d->polygons[i]->points[j].vect->y < ymax->y )
				ymax = d->polygons[i]->points[j].vect;
			if ( d->polygons[i]->points[j].vect->z < zmin->z )
				zmin = d->polygons[i]->points[j].vect;
			if ( d->polygons[i]->points[j].vect->z < zmax->z )
				zmax = d->polygons[i]->points[j].vect;
		}

	x = (xmax->x - xmin->x)*(xmax->x - xmin->x) + 
	    (xmax->y - xmin->y)*(xmax->y - xmin->y) +
	    (xmax->z - xmin->z)*(xmax->z - xmin->z);
	y = (ymax->x - ymin->x)*(ymax->x - ymin->x) + 
	    (ymax->y - ymin->y)*(ymax->y - ymin->y) +
	    (ymax->z - ymin->z)*(ymax->z - ymin->z);
	z = (zmax->x - zmin->x)*(zmax->x - zmin->x) + 
	    (zmax->y - zmin->y)*(zmax->y - zmin->y) +
	    (zmax->z - zmin->z)*(zmax->z - zmin->z);
	
	/* maximum dist */
	if ( x > y ) {
		maxdist = x;
		d1 = xmin;	d2 = xmax;
	}
	else {
		maxdist = y;
		d1 = ymin;	d2 = ymax;
	}
	if ( z > maxdist ) {
		maxdist = z;
		d1 = zmin;	d2 = zmax;
	}

	initV( &(result.center), (d1->x + d2->x)/2, 
			(d1->y + d2->y)/2, (d1->z + d2->z)/2 );
	radsq = (d2->x - result.center.x)*(d2->x - result.center.x) +
		(d2->y - result.center.y)*(d2->y - result.center.y) +
		(d2->z - result.center.z)*(d2->z - result.center.z);
	result.radius = sqrt(x);
	
	for ( i = 0; i < d->npoly; i++ )
		for ( j = 0; j < d->polygons[i]->npoints; j++ ) {
			a= (d->polygons[i]->points[j].vect->x-result.center.x)*
			   (d->polygons[i]->points[j].vect->x-result.center.x)+
			   (d->polygons[i]->points[j].vect->y-result.center.y)*
			   (d->polygons[i]->points[j].vect->y-result.center.y)+
			   (d->polygons[i]->points[j].vect->z-result.center.z)*
			   (d->polygons[i]->points[j].vect->z-result.center.z);
			if ( a > radsq ) {
				b = sqrt( x );
				result.radius = (result.radius + b)/2;
				radsq = result.radius*result.radius;
				a = b - result.radius;
				initV( &(result.center), 
   (result.radius*result.center.x + a*d->polygons[i]->points[j].vect->x)/b,
   (result.radius*result.center.y + a*d->polygons[i]->points[j].vect->y)/b,
   (result.radius*result.center.z + a*d->polygons[i]->points[j].vect->z)/b );
			}
		}
	return result;
}


/* example: 3, False, color1, x, y, z, color2, ...*/
#ifdef CLIENT
POLY_UL *createPolyD ( sInt n, Bool onecolor, ... ) {
	ggi_color	color;
#else
POLY_UL *createPolyD ( sInt n, ... ) {
#endif
	POLY_UL		*newpoly;
	VECTOR_UL	p1, p2;
	va_list		vl;
	sInt		i;

	if ( n < 3 ) {
		fprintf( stderr, "\nToo few points at PolyD" );
		return NULL;
	}
	newpoly = (POLY_UL *)malloc(sizeof(POLY_UL));
	if ( !newpoly ) {
		fprintf( stderr, "\n%s PolyD", ALLOCERROR );
		return NULL;
	}
	newpoly->points = (POINT_UL *)malloc(n*sizeof(POINT_UL));
	if ( !newpoly->points ) {
		fprintf( stderr, "\n%s PolyD, 2", ALLOCERROR );
		return NULL;
	}
	for ( i = 0; i < n; i++ ) {
		newpoly->points[i].vect = (VECTOR_UL *)malloc(sizeof(VECTOR_UL));
		if ( !newpoly->points[i].vect ) {
			fprintf( stderr, "\n%s PolyD, 3.1", ALLOCERROR );
			return NULL;
		}
		newpoly->points[i].norm = 
				(VECTOR_UL *)malloc( sizeof(VECTOR_UL) );
		if ( !newpoly->points[i].norm ) {
			fprintf( stderr, "\n%s PolyD, 3.2", ALLOCERROR );
			return NULL;
		}
	}

	newpoly->npoints = n;

#ifdef CLIENT
	va_start( vl, onecolor );
	newpoly->color = (ggi_color *)malloc(n*sizeof(ggi_color));
	if ( !newpoly->color ) {
		fprintf( stderr, "\n%s PolyD, 5", ALLOCERROR );
		return NULL;
	}

	if ( onecolor == True ) {
		newpoly->onecolor = True;
		copyColor( &color, &va_arg( vl, ggi_color ) );
		for ( i = 0; i < n; i++ )
			copyColor( &newpoly->color[i], &color );
	}
	else
		newpoly->onecolor = False;
#else
	va_start( vl, n );
#endif
	for ( i = 0; i < n; i++ ) {
#ifdef CLIENT		
		if ( !newpoly->onecolor ) {
			copyColor( &newpoly->color[i], &va_arg( vl, ggi_color ) );
		}
#endif
		newpoly->points[i].vect->x = va_arg( vl, dFloat );
		newpoly->points[i].vect->y = va_arg( vl, dFloat );
		newpoly->points[i].vect->z = va_arg( vl, dFloat );
	}
	pointstoV( &newpoly->points[1], &newpoly->points[0], &p1 );
	pointstoV( &newpoly->points[2], &newpoly->points[0], &p2 );
	crossprodV( &p2, &p1, &(newpoly->normal) );
	normalizeV( &(newpoly->normal) );
	centerP( newpoly, &(newpoly->center) );
	va_end( vl );
	return newpoly;
}

#ifdef CLIENT
POLY_UL *createPolyP ( sInt n, Bool onecolor, ggi_color *color, POINT_UL *points ) {
#else
POLY_UL *createPolyP ( sInt n, POINT_UL *points ) {
#endif
	POLY_UL *newpoly;
	VECTOR_UL p1, p2;
	sInt i;
	
	newpoly = (POLY_UL *)malloc(sizeof(POLY_UL));
	if ( !newpoly ) {
		fprintf( stderr, "\n%s PolyP", ALLOCERROR );
		return NULL;
	}

	newpoly->points = (POINT_UL *)malloc(n*sizeof(POINT_UL));
	if ( !newpoly->points ) {
		fprintf( stderr, "\n%s PolyP, 2", ALLOCERROR );
		return NULL;
	}

	for ( i = 0; i < n; i++ ) {
		if ( !(newpoly->points[i].vect = (VECTOR_UL *)malloc(sizeof(VECTOR_UL)))) {
			fprintf( stderr, "\n%s PolyP, 3", ALLOCERROR );
			return NULL;
		}
		copyV( newpoly->points[i].vect, points[i].vect );
		if ( points[i].norm ) {
			if ( !(newpoly->points[i].norm = (VECTOR_UL *)malloc(sizeof(VECTOR_UL)))) {
				fprintf( stderr, "\n%s PolyP, 4", ALLOCERROR );
				return NULL;
			}
			copyV( newpoly->points[i].norm, points[i].norm );
		}
#ifdef CLIENT
		newpoly->points[i].u = points[i].u;
		newpoly->points[i].v = points[i].v;
#endif

	}
	newpoly->npoints = n;
#ifdef CLIENT
	newpoly->color = (ggi_color *)malloc(n*sizeof(ggi_color));
	if ( !newpoly->color ) {
		fprintf( stderr, "\n%s PolyD, 4", ALLOCERROR );
		return NULL;
	}
	if ( onecolor == True ) {
		newpoly->onecolor = True;
		copyColor( &newpoly->color[0], &color[0] );
		for ( i = 1; i < n; i++ )
			copyColor( &newpoly->color[i], &newpoly->color[0] );
	}
	else {
		newpoly->onecolor = False;
		for ( i = 0; i < n; i++ )
			copyColor( &newpoly->color[i], &color[i] );
	}
#endif	
	pointstoV( &newpoly->points[1], &newpoly->points[0], &p1 );
	pointstoV( &newpoly->points[2], &newpoly->points[0], &p2 );
	crossprodV( &p2, &p1, &(newpoly->normal) );
	normalizeV( &(newpoly->normal) );
	centerP( newpoly, &(newpoly->center) );
	return newpoly;
}

#ifdef CLIENT
POLY_UL *createPolyPV ( sInt n, Bool onecolor, ggi_color *color, ... ) {
#else
POLY_UL *createPolyPV ( sInt n, ... ) {
#endif
	POLY_UL		*newpoly;
	VECTOR_UL	p1, p2;
	POINT_UL	*pp;
	sInt		i;
	va_list		vl;
	
	newpoly = (POLY_UL *)malloc(sizeof(POLY_UL));
	if ( !newpoly ) {
		fprintf( stderr, "\n%s PolyP", ALLOCERROR );
		return NULL;
	}
	newpoly->npoints = n;
#ifdef CLIENT
	newpoly->color = (ggi_color *)malloc(n*sizeof(ggi_color));
	if ( !newpoly->color ) {
		fprintf( stderr, "\n%s PolyD, 4", ALLOCERROR );
		return NULL;
	}
	if ( onecolor == True ) {
		newpoly->onecolor = True;
		copyColor( &newpoly->color[0], &color[0] );
		for ( i = 1; i < n; i++ )
			copyColor( &newpoly->color[i], &newpoly->color[0] );
	}
	else {
		newpoly->onecolor = False;
		for ( i = 0; i < n; i++ )
			copyColor( &newpoly->color[i], &color[i] );
	}
	va_start( vl, color );
#else
	va_start( vl, n );
#endif
	newpoly->points = (POINT_UL *)malloc(n*sizeof(POINT_UL));
	if ( !newpoly->points ) {
		fprintf( stderr, "\n%s PolyD, 5", ALLOCERROR );
		return NULL;
	}
	for ( i = 0; i < n; i++ ) {
		pp = &va_arg( vl, POINT_UL );
		newpoly->points[i].vect = pp->vect;
		fprintf( stderr, "%f %f %f\n", pp->vect->x, pp->vect->y, pp->vect->z );
		fflush( stderr );
		newpoly->points[i].norm = pp->norm;
#ifdef CLIENT
		newpoly->points[i].u = pp->u;
		newpoly->points[i].v = pp->v;
#endif
	}

	pointstoV( &newpoly->points[1], &newpoly->points[0], &p1 );
	pointstoV( &newpoly->points[2], &newpoly->points[0], &p2 );
	crossprodV( &p2, &p1, &(newpoly->normal) );
	normalizeV( &(newpoly->normal) );
	centerP( newpoly, &(newpoly->center) );
	va_end( vl );
	return newpoly;
}

POLY_UL *copyPoly ( POLY_UL *p ) {
	POINT_UL	*newpoints;
	sInt		i;
	
	newpoints = (POINT_UL *)malloc(p->npoints*sizeof(POINT_UL));
	if ( !newpoints ) {
		fprintf( stderr, "\n%s copyPoly", ALLOCERROR );
		return NULL;
	}
	for ( i = 0; i < p->npoints; i++ ) {
		if ( !(newpoints[i].vect = (VECTOR_UL *)malloc(sizeof(VECTOR_UL))) ) {
			fprintf( stderr, "\n%s copyPoly 2", ALLOCERROR );
			return NULL;
		}
		copyV( newpoints[i].vect, p->points[i].vect );
	}
#ifdef CLIENT
	return(createPolyP( p->npoints, (p->onecolor == True ? True : False ),
			p->color, newpoints ));
#else
	return(createPolyP( p->npoints, newpoints ));
#endif
	free( newpoints );
}

void freePoly ( POLY_UL *p ) {
	sInt 		i;

#ifdef CLIENT
	free( p->color );
	if ( p->flatcolor )
		free( p->flatcolor );
#endif
	for ( i = 0; i < p->npoints; i++ ) {
		if ( p->points[i].vect )
			free( p->points[i].vect );
		if ( p->points[i].norm )
			free( p->points[i].norm );
	}
	free( p->points );
	free( p );
}

/* noftri = npoints -2 */
#ifdef CLIENT
TRISTRIP_UL *createTriStripD ( sInt noftri, ggi_color color, ... ) {
#else
TRISTRIP_UL *createTriStripD ( sInt noftri, ... ) {
#endif
	TRI_UL		**strip;
	POINT_UL	pst[noftri-1];
	VECTOR_UL	p[3];
	va_list		vl;
	sInt		i;
	
	strip = (TRI_UL **)malloc(noftri*sizeof(TRI_UL *));
	if ( !strip ) {
		fprintf( stderr, "\n%s TriStripD", ALLOCERROR );
		return NULL;
	}
	
#ifdef CLIENT
	va_start( vl, color );
#else
	va_start( vl, noftri );
#endif
	p[0].x = va_arg( vl, dFloat );
	p[0].y = va_arg( vl, dFloat );
	p[0].z = va_arg( vl, dFloat );
	p[1].x = va_arg( vl, dFloat );
	p[1].y = va_arg( vl, dFloat );
	p[1].z = va_arg( vl, dFloat );
	p[2].x = va_arg( vl, dFloat );
	p[2].y = va_arg( vl, dFloat );
	p[2].z = va_arg( vl, dFloat );
#ifdef CLIENT
	strip[0] = createPolyD( 3, True, color, p[0].x, p[0].y, p[0].z, p[1].x, 
			p[1].y, p[1].z, p[2].x, p[2].y, p[2].z );
#else
	strip[0] = createPolyD( 3, p[0].x, p[0].y, p[0].z, p[1].x, p[1].y,
			p[1].z, p[2].x, p[2].y, p[2].z );
#endif	
	
	for ( i = 0; i < noftri-1; i++ ) {
		pst[i].vect = (VECTOR_UL *)malloc(sizeof(VECTOR_UL));
		if ( !pst[i].vect ) {
			fprintf( stderr, "\n%s StripD", ALLOCERROR );
			return NULL;
		}
		/* the normal will not be used... */
		pst[i].vect->x = va_arg( vl, dFloat );
		pst[i].vect->y = va_arg( vl, dFloat );
		pst[i].vect->z = va_arg( vl, dFloat );
#ifdef CLIENT
		strip[i+1] = createPolyPV( 3, True, &color,
				strip[i]->points[1].vect, 
				strip[i]->points[2].vect, pst[i].vect );
#else
		strip[i+1] = createPolyPV( 3, strip[i]->points[1].vect, 
				strip[i]->points[2].vect, pst[i].vect );
#endif
	}
	va_end( vl );
/* BUG? check pointers here, the revision wasnt careful */
	return (TRISTRIP_UL *)strip;
}

#ifdef CLIENT
TRISTRIP_UL *createTriStripP ( sInt noftri, ggi_color color, ... ) {
#else
TRISTRIP_UL *createTriStripP ( sInt noftri, ... ) {
#endif
	TRI_UL		**strip;
	POINT_UL	**p;
	va_list 	vl;
	sInt 		i;

	strip = (TRI_UL **)malloc(noftri*sizeof(TRI_UL *));
	if ( !strip ) {
		fprintf( stderr, "\n%s TriStripP", ALLOCERROR );
		return NULL;
	}
	
	p = (POINT_UL **)malloc(3*sizeof(TRISTRIP_UL *));
	if ( !p ) {
		fprintf( stderr, "\n%s TriStripP, 2", ALLOCERROR );
		return NULL;
	}
	
#ifdef CLIENT
	va_start( vl, color );
#else
	va_start( vl, noftri );
#endif
	p[0] = va_arg( vl, POINT_UL *);
	p[1] = va_arg( vl, POINT_UL *);
	p[2] = va_arg( vl, POINT_UL *);
#ifdef CLIENT
	strip[0] = createPolyPV( 3, True, &color, p[0], p[1], p[2] );
#else
	strip[0] = createPolyPV( 3, p[0], p[1], p[2] );
#endif
	
	for ( i = 2; i < noftri-1; i++ ) {
#ifdef CLIENT
		strip[i+1] = createPolyPV( 3, True, &color, strip[i]->points[1],
				strip[i]->points[2], va_arg( vl, POINT_UL ) );
#else
		strip[i+1] = createPolyPV( 3, strip[i]->points[1],
				strip[i]->points[2], &va_arg( vl, POINT_UL ) );
#endif
	}
	va_end( vl );
	free( p );
	return strip;
}

#ifdef CLIENT
TRISTRIP_UL *createTriStripColorD ( sInt noftri, ... ) {
	TRISTRIP_UL *strip;
	POINT_UL pst[noftri-1];
	VECTOR_UL p[3];
	ggi_color color;
	va_list vl;
	sInt i;
	
	strip = (TRISTRIP_UL *)malloc(noftri*sizeof(TRISTRIP_UL));
	if ( !strip ) {
		fprintf( stderr, "\n%s TriStripColorD", ALLOCERROR );
		return NULL;
	}
		
	va_start( vl, noftri );
	p[0].x = va_arg( vl, dFloat );
	p[0].y = va_arg( vl, dFloat );
	p[0].z = va_arg( vl, dFloat );
	p[1].x = va_arg( vl, dFloat );
	p[1].y = va_arg( vl, dFloat );
	p[1].z = va_arg( vl, dFloat );
	p[2].x = va_arg( vl, dFloat );
	p[2].y = va_arg( vl, dFloat );
	p[2].z = va_arg( vl, dFloat );
	copyColor( &color, &va_arg( vl, ggi_color ) );
	strip[0] = createPolyD( 3, True, color, p[0].x, p[0].y,
			p[0].z,	p[1].x, p[1].y, p[1].z, p[2].x, p[2].y, p[2].z );
	
	for ( i = 0; i < noftri-1; i++ ) {
		pst[i].vect = (VECTOR_UL *)malloc(sizeof(VECTOR_UL));
		if ( !pst[i].vect ) {
			fprintf( stderr, "\n%s StripColorD", ALLOCERROR );
			return NULL;
		}
		pst[i].vect->x = va_arg( vl, dFloat );
		pst[i].vect->y = va_arg( vl, dFloat );
		pst[i].vect->z = va_arg( vl, dFloat );
		copyColor( &color, &va_arg( vl, ggi_color ) );
		strip[i+1] = createPolyPV( 3, True, &color, 
			strip[i]->points[1].vect, 
			strip[i]->points[2].vect, pst[i].vect );
	}
	va_end( vl );
	return strip;
}

TRISTRIP_UL *createTriStripColorP ( sInt noftri, ... ) {
	TRI_UL **strip;
	POINT_UL **p;
	va_list vl;
	sInt i;

	strip = (TRI_UL **)malloc(noftri*sizeof(TRI_UL *));
	if ( !strip ) {
		fprintf( stderr, "\n%s TriStrip", ALLOCERROR );
		return NULL;
	}
	
	p = (POINT_UL **)malloc(3*sizeof(TRISTRIP_UL *));
	if ( !p ) {
		fprintf( stderr, "\n%s TriStrip, 2", ALLOCERROR );
		return NULL;
	}
	
	va_start( vl, noftri );
	p[0] = va_arg( vl, POINT_UL *);
	p[1] = va_arg( vl, POINT_UL *);
	p[2] = va_arg( vl, POINT_UL *);
	strip[0] = (TRI_UL *)createPolyPV( 3, False, &va_arg( vl, ggi_color ), 
			p[0], p[1], p[2] );
	
	for ( i = 1; i < noftri; i++ ) {
		strip[i] = (TRI_UL *)createPolyPV( 3, False, 
			&va_arg( vl, ggi_color ), strip[i-1]->points[1].vect,
			strip[i-1]->points[2].vect, va_arg( vl, POINT_UL *) );
	}
	va_end( vl );
	free( p );
	return strip;
}
#endif

VECTOR_UL id = { 1, 0, 0 };

DOODAD_UL *createDoodad ( uByte *filename ) {
	DOODAD_UL	*tmp, *dlast;
	sInt		i, j;
	
	tmp = (DOODAD_UL *)malloc(sizeof(DOODAD_UL));
	if ( !tmp ) {
		fprintf( stderr, "%s createDoodad 1\n", ALLOCERROR );
		return NULL;
	}

	if ( readGeometry( filename, tmp ) != OK ) {
		fprintf( stderr, "Check %s, it's corrupt!!!\n", filename );
		free( tmp );
		return NULL;
	}

	/* set local system equal world system */
	zeroV( &tmp->world );
	copyV( &tmp->dir, &id );
	tmp->angle = 0;

	for ( i = 0; i < 3; i++ )
		for ( j = 0; j < 3; j++ )
			tmp->localtoworld[i][j] =
				tmp->worldtolocal[i][j] = ( i == j ? 1 : 0 );
	
	if ( !firstdoodad )
		firstdoodad = tmp;
	else {
		for ( dlast = firstdoodad; dlast->next != NULL; dlast = dlast->next ) ;
		dlast->next = tmp;
	}
	tmp->next = (DOODAD_UL *)NULL;
	tmp->ncopies = 0;
	return tmp;
}

/* LAST poly should be NULL!!!!!! */
DOODAD_UL *createDoodadP ( uByte *name, POLY_UL **p ) {
	DOODAD_UL	*tmp, *dlast;
	sInt		i, j;
	
	tmp = (DOODAD_UL *)malloc(sizeof(DOODAD_UL));
	if ( !tmp ) {
		fprintf( stderr, "%s createDoodad 1\n", ALLOCERROR );
		return NULL;
	}
	
	if ( name != NULL )
		tmp->name = strdup( name );

	for ( tmp->npoly = 0; p[tmp->npoly]; tmp->npoly++ ) ;

	tmp->polygons = (POLY_UL **)malloc(tmp->npoly*sizeof(POLY_UL *));

	for ( i = 0; i < tmp->npoly; i++ )
		tmp->polygons[i] = p[i];
	/* position in 0,0,0, and orient towards x axis */
	zeroV( &tmp->world );
	copyV( &tmp->dir, &id );
	tmp->angle = 0;

	/* set local system equal world system */
	for ( i = 0; i < 3; i++ )
		for ( j = 0; j < 3; j++ )
			tmp->localtoworld[i][j] =
				tmp->worldtolocal[i][j] = ( i == j ? 1 : 0 );
	
	tmp->sphere = circSphere( tmp );
	centroid( tmp, &tmp->centroid );

	if ( !firstdoodad )
		firstdoodad = tmp;
	else {
		for ( dlast = firstdoodad; dlast->next != NULL; dlast = dlast->next ) ;
		dlast->next = tmp;
	}
	tmp->next = (DOODAD_UL *)NULL;
	tmp->ncopies = 0;
	return tmp;
}


DOODAD_UL *copyDoodad ( DOODAD_UL *src ) {
	DOODAD_UL	*tmp, *dlast;
	sInt		i, j;
	
	if ( !src ) {
		fprintf( stderr, "src is NULL\n" );
		return NULL;
	}
	tmp = (DOODAD_UL *)malloc(sizeof(DOODAD_UL));
	if ( !tmp ) {
		fprintf( stderr, "%s copyDoodad\n", ALLOCERROR );
		return NULL;
	}
	tmp->polygons = (POLY_UL **)malloc(src->npoly*sizeof(POLY_UL *));
	for ( i = 0; i < src->npoly; i++ )
		tmp->polygons[i] = copyPoly( src->polygons[i] );
	tmp->npoly = src->npoly;
	tmp->nvert = src->nvert;

	/* set local system equal world system */
	for ( i = 0; i < 3; i++ )
		for ( j = 0; j < 3; j++ ) {
			tmp->localtoworld[i][j] = src->localtoworld[i][j];
			tmp->worldtolocal[i][j] = src->worldtolocal[i][j];
		}

	copyV( &tmp->sphere.center, &src->sphere.center );
	copyV( &tmp->centroid, &src->centroid );
	tmp->sphere.radius = src->sphere.radius;
	copyV( &(tmp->world), &(src->world) );
	copyV( &(tmp->dir), &(src->dir) );
	tmp->ncopies = 0;
	if ( !firstdoodad )
		firstdoodad = tmp;
	else {
		for ( dlast = firstdoodad; dlast->next != NULL; dlast = dlast->next ) ;
		dlast->next = tmp;
	}
	tmp->next = (DOODAD_UL *)NULL;
	tmp->ncopies = 0;
	return tmp;
}

/* copies everything, but geometry is shared */
DOODAD_UL *copySharedDoodad ( DOODAD_UL *src ) {
	DOODAD_UL 	*tmp, *dlast;
	sInt		i, j;
		
	if ( !src ) {
		fprintf( stderr, "src (%s) is NULL: copySharedDoodad\n", 
				src->name );
		return NULL;
	}
	
	for ( dlast = firstdoodad; dlast != src; dlast = dlast->next ) ;
	if ( !dlast ) {
		fprintf( stderr, "src (%s) isn't in list: copySharedDoodad\n", 
				src->name );
		return NULL;
	}

	tmp = (DOODAD_UL *)malloc(sizeof(DOODAD_UL));
	if ( !tmp ) {
		fprintf( stderr, "\n%s copyDoodad", ALLOCERROR );
		return NULL;
	}
	tmp->npoly = src->npoly;
	tmp->nvert = src->nvert;
	copyV( &(tmp->world), &(src->world) );
	copyV( &(tmp->dir), &(src->dir) );

	/* set local system equal world system */
	for ( i = 0; i < 3; i++ )
		for ( j = 0; j < 3; j++ ) {
			tmp->localtoworld[i][j] = src->localtoworld[i][j];
			tmp->worldtolocal[i][j] = src->worldtolocal[i][j];
		}

	tmp->polygons = src->polygons;
	tmp->ncopies = 0;
	copyV( &tmp->sphere.center, &src->sphere.center );
	copyV( &tmp->centroid, &src->centroid );
	tmp->sphere.radius = src->sphere.radius;
	for ( dlast = firstdoodad; dlast->next != NULL; dlast = dlast->next ) ;
	dlast->next = tmp;
	src->ncopies++;
	return tmp;
}

void freeDoodad ( DOODAD_UL *dood ) {
	DOODAD_UL	*tmp;
	sInt		i;

	if ( dood->ncopies ) {
		tmp = dood;
		do {
		 	tmp = tmp->next;
			if ( tmp->polygons == dood->polygons ) {
				tmp->ncopies = dood->ncopies - 1;
				break;
			}
		} while ( tmp->next != NULL );
	}
	else {
		for ( i = 0; i < dood->npoly; i++ )
			freePoly( dood->polygons[i] );
		free( dood->polygons );
	}
	tmp = firstdoodad;
	
	if ( tmp == dood )
		firstdoodad = dood->next;
	else while ( tmp->next != dood )
		tmp = tmp->next;
	
	tmp->next = dood->next;
	
	if ( dood->name )
		free( dood->name );
	free ( dood );
}
