/* makeulaw.c */

/* This file contains a the source code for a tiny program which generates
 * a table for converting u-Law samples to 16-bit linear samples, or
 * (when invoked with "-l") vice-versa.
 */

/* convert u-Law byte 'b' to a linear word */
static int mu2lin(int b)
{
	int mu = b ^ 0xff;         /* flip all 8 bits   */
	int s = (mu & 0x80) >> 7 ; /* X------- sign bit */
	int e = (mu & 0x70) >> 4 ; /* -XXX---- exponent */
	int f = (mu & 0x0F) >> 0 ; /* ----XXXX mantissa */
	int lin;

	lin = ((f<<(e+1)) - 0x20 + (0x20 << e));
	return (s==1)?-lin:lin;
}

/* convert linear word 'w' to a u-Law byte */
static char lin2mu(int w)
{
	int	closest, closediff, diff, b;

	for (b = 0; b < 256; b++)
	{
		diff = mu2lin(b) - w;
		if (diff < 0)
			diff = -diff;
		if (b == 0 || diff < closediff)
			closest = b, closediff = diff;
	}
	return (char)closest;
}

main(int argc, char **argv)
{
	int	b, w;

	printf("/* automatically generated by makeulaw.c */\n");
	if (argc == 1)
	{
		printf("\nstatic short ulaw2linear[256] = {");
		for (b = 0; b < 256; b++)
			printf("%s%d%s",
				(b % 9 == 0)? "\n\t" : " ",
				mu2lin(b),
				b == 255 ? "" : ",");
		printf("\n};\n");
	}
	else
	{
		printf("\nstatic char linear2ulaw[4096] = {");
		for (w = 0; w < 4096; w++)
			printf("%s%d%s",
				(w % 12 == 0)? "\n\t" : " ",
				lin2mu(w & 2048 ? w - 4096 : w),
				w == 4095 ? "" : ",");
		printf("\n};\n");
	}

	exit(0);
}
